package be.iminds.ilabt.jfed.lowlevel.api.test;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.ServerBuilder;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.ServiceBuilder;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.api.RestApi;
import be.iminds.ilabt.jfed.lowlevel.connection.HttpConnection;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedException;
import be.iminds.ilabt.jfed.lowlevel.connection_pool.JFedConnectionProvider;
import be.iminds.ilabt.jfed.lowlevel.lib.*;
import be.iminds.ilabt.jfed.lowlevel.testbed_info.ApiInfo;
import be.iminds.ilabt.jfed.lowlevel.testbed_info.TestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.testing.base.*;
import be.iminds.ilabt.jfed.lowlevel.authority.legacy.TargetAuthority;
import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.dropwizard.jackson.Jackson;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;
import java.util.stream.Collectors;

/**
 * A Test of GPULab
 */
public class TestGPULab extends BaseApiTest<GPULabTestConfig> {
    @Inject
    public TestGPULab(@Nonnull Logger logger,
                      @Nullable TargetAuthority testedAuthority,
                      @Nonnull GeniUserProvider geniUserProvider,
                      @Nonnull GPULabTestConfig testConfig,
                      @Nonnull JFedConnectionProvider connectionProvider,
                      @Nonnull TestbedInfoSource testbedInfoSource,
                      @Nonnull JFedPreferences jFedPreferences) {
        super(logger, testedAuthority, geniUserProvider, testConfig,
                connectionProvider, testbedInfoSource,
                jFedPreferences);
    }


    private static final ObjectMapper MAPPER = Jackson.newObjectMapper();
    private static final ApiTestMetaData metadata = new ApiTestMetaData<GPULabTestConfig>() {
        @Nonnull
        @Override
        public String getTestDescription() {
            return "GPULab Test: tests connectivity and basic functionality";
        }

        @Nonnull
        @Override
        public Class<GPULabTestConfig> getConfigClass() {
            return GPULabTestConfig.class;
        }

        @Nonnull
        @Override
        public GPULabTestConfig parseApiTestConfig(@Nonnull String apiTestConfigString) {
            if (apiTestConfigString == null || apiTestConfigString.trim().isEmpty())
                return new GPULabTestConfig();
            try {
                return MAPPER.readValue(apiTestConfigString, GPULabTestConfig.class);
            } catch (IOException e) {
                throw new RuntimeException("Invalid Test config", e);
            }
        }

        @Nonnull
        @Override
        public ApiTestConfigEditor<GPULabTestConfig> getApiTestConfigEditor() {
            return new JsonTestConfigEditor(GPULabTestConfig.class);
        }

    };
    public static ApiTestMetaData getMetaData() {
        return metadata;
    }


    private URL url;
    public void setUp() throws MalformedURLException {
        url = new URL(testConfig.getGpulabURL());

        note("done");
    }


    @Test(groups = {"test"})
    public void testGPULab() throws JFedException {
        GPULabApi api = new GPULabApi(logger, getJFedPreferences());
        ServiceBuilder serviceBuilder = new ServiceBuilder()
                .setApi(api.getApi().getName().getId())
                .setApiVersion(api.getApi().getVersion())
                .setUrl(testConfig.getGpulabURL())
                .setId(Integer.MAX_VALUE-5); //dummy ID
        Server server = new ServerBuilder()
                .setName("GPULab")
                .setBaseUrl(testConfig.getGpulabURL())
                .addServiceBuilder(serviceBuilder)
                .setId(Integer.MAX_VALUE-6) //dummy ID
                .create();
        JFedConnection con = connectionProvider.getConnectionByAuthority(user, server, api.getApi());

        List<GPULabCluster> clusters = api.getClusters((HttpConnection) con);
        assertNotEmpty(clusters, "There are no active GPULab clusters");
        note("GPULab version: "+clusters.get(0).getVersion());

        String clusterIds = clusters.stream().map(GPULabCluster::getId).collect(Collectors.joining(", "));
        note("Clusters: "+clusterIds);
        
        int allGpuCount = clusters.stream().map(GPULabCluster::getGpuTotalCount).filter(Objects::nonNull).mapToInt(Integer::intValue).sum();
        int freeGpuCount = clusters.stream().map(GPULabCluster::getGpuAvailableCount).filter(Objects::nonNull).mapToInt(Integer::intValue).sum();
        note("Total GPUs: "+allGpuCount);
        note("Free  GPUs: "+freeGpuCount);
        
        int allCpuCount = clusters.stream().map(GPULabCluster::getCpuTotalCount).filter(Objects::nonNull).mapToInt(Integer::intValue).sum();
        int freeCpuCount = clusters.stream().map(GPULabCluster::getCpuAvailableCount).filter(Objects::nonNull).mapToInt(Integer::intValue).sum();
        note("Total Cpus: "+allCpuCount);
        note("Free  Cpus: "+freeCpuCount);

        List<GPULabJob> jobs = api.getJobs((HttpConnection) con);
        note("Job count: "+jobs.size());
        note("Jobs: "+jobs.stream().map(GPULabJob::getUuid).collect(Collectors.joining(", ")));

        List<GPULabSlave> slaves = api.getSlaves((HttpConnection) con);
        assertNotEmpty(slaves, "There are no active GPULab slaves");

        String slaveHostnames = slaves.stream().map(GPULabSlave::getSlaveHostname).filter(Objects::nonNull).collect(Collectors.joining(", "));
        note("Slaves: "+slaveHostnames);

        addExtraResult(new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                .setKey("clusters")
                .setLarge(false)
                .setHttpMediaTypeToText()
                .setValue(clusters.size())
        );
        addExtraResult(new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                .setKey("slaves")
                .setLarge(false)
                .setHttpMediaTypeToText()
                .setValue(slaves.size())
        );
        addExtraResult(new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                .setKey("cluster_ids")
                .setLarge(false)
                .setHttpMediaTypeToText()
                .setValue(clusterIds)
        );
        addExtraResult(new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                .setKey("slave_hostnames")
                .setLarge(false)
                .setHttpMediaTypeToText()
                .setValue(slaveHostnames)
        );
        addExtraResult(new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                .setKey("recent_job_count")
                .setLarge(false)
                .setHttpMediaTypeToText()
                .setValue(jobs.size())
        );
        addExtraResult(new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                .setKey("gpu")
                .setLarge(false)
                .setHttpMediaTypeToText()
                .setValue(freeGpuCount+"/"+allGpuCount)
        );
        addExtraResult(new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                .setKey("cpu")
                .setLarge(false)
                .setHttpMediaTypeToText()
                .setValue(freeCpuCount+"/"+allCpuCount)
        );

        assertTrue(allGpuCount >= testConfig.getMinGPUs(),
                "Too few GPU's in cluster -> minimum expected="+testConfig.getMinGPUs()+" actual="+allGpuCount);

        assertTrue(allCpuCount >= testConfig.getMinCPUs(),
                "Too few CPU's in cluster -> minimum expected="+testConfig.getMinCPUs()+" actual="+allCpuCount);

        assertTrue(clusters.size() >= testConfig.getMinClusters(),
                "Cluster count too small -> minimum expected="+testConfig.getMinCPUs()+" actual="+clusters.size());

        assertTrue(slaves.size() >= testConfig.getMinSlaves(),
                "Slave count too small -> minimum expected="+testConfig.getMinSlaves()+" actual="+slaves.size());

        note("Test completed successfully.");
    }



    public static class GPULabApi extends RestApi {
        public GPULabApi(@Nonnull Logger logger, @Nonnull JFedPreferences jFedPreferences) {
            super(logger, new ApiInfo.Api(ApiInfo.ApiName.GENERIC_HTTPS_CLIENTAUTH, 1), jFedPreferences);
        }

        @Nonnull
        @Override
        public String getName() {
            return getApiName();
        }

        /**
         * A human readable name for the implemented API
         *
         * @return "SLA Collector";
         */
        static public String getApiName() {
            return "GPULabApi";
        }

        @ApiMethod(order = 1, hint = "GET gpulab/api/v2.0/clusters")
        public List<GPULabCluster> getClusters(@Nonnull HttpConnection con) throws JFedException {
            String javaName = "getClusters";
            String relUrl = "gpulab/api/v2.0/clusters";
            Map<String, Object> urlParams = makeMethodParameters();

            RestApi.RestReply<String> r = internalGet(con, javaName, relUrl, urlParams, urlParams);
            try {
                return MAPPER.readValue(r.getValue(), new TypeReference<List<GPULabCluster>>() {});
            } catch (IOException e) {
                throw new JFedException("Exception mapping reply", e);
            }
        }

        @ApiMethod(order = 2, hint = "GET gpulab/api/v2.0/slaves")
        public List<GPULabSlave> getSlaves(@Nonnull HttpConnection con) throws JFedException {
            String javaName = "getSlaves";
            String relUrl = "gpulab/api/v2.0/slaves";
            Map<String, Object> urlParams = makeMethodParameters();

            RestApi.RestReply<String> r = internalGet(con, javaName, relUrl, urlParams, urlParams);
            try {
                return MAPPER.readValue(r.getValue(), new TypeReference<List<GPULabSlave>>() {});
            } catch (IOException e) {
                throw new JFedException("Exception mapping reply", e);
            }
        }

        @ApiMethod(order = 3, hint = "GET gpulab/api/v2.0/jobs?max_hours=900&max_count=100")
        public List<GPULabJob> getJobs(@Nonnull HttpConnection con) throws JFedException {
            String javaName = "getJobs";
            String relUrl = "gpulab/api/v2.0/jobs?max_hours=900&max_count=100";
            Map<String, Object> urlParams = makeMethodParameters();

            RestApi.RestReply<String> r = internalGet(con, javaName, relUrl, urlParams, urlParams);
            try {
                return MAPPER.readValue(r.getValue(), new TypeReference<List<GPULabJob>>() {});
            } catch (IOException e) {
                throw new JFedException("Exception mapping reply", e);
            }
        }
    }

    @JsonInclude(JsonInclude.Include.NON_NULL)
    @JsonIgnoreProperties(ignoreUnknown=true)
    private static class GPULabCluster {
        @Nullable private final String id;
        @Nullable private final String comment;
        @Nullable private final Integer cpuAvailableCount;
        @Nullable private final Integer cpuTotalCount;
        @Nullable private final Integer gpuAvailableCount;
        @Nullable private final Integer gpuTotalCount;
        @Nullable private final List<String> gpuTypes;
        @Nullable private final Integer slaveCount;
        @Nullable private final String version;

        @JsonCreator
        public GPULabCluster(@Nullable @JsonProperty("cluster_id") String id,
                             @Nullable @JsonProperty("comment") String comment,
                             @Nullable @JsonProperty("cpu_available") Integer cpuAvailableCount,
                             @Nullable @JsonProperty("cpu_count") Integer cpuTotalCount,
                             @Nullable @JsonProperty("gpu_available") Integer gpuAvailableCount,
                             @Nullable @JsonProperty("gpu_count") Integer gpuTotalCount,
                             @Nullable @JsonProperty("gpu_models") List<String> gpuTypes,
                             @Nullable @JsonProperty("slave_count") Integer slaveCount,
                             @Nullable @JsonProperty("version") String version) {
            this.id = id;
            this.comment = comment;
            this.cpuAvailableCount = cpuAvailableCount;
            this.cpuTotalCount = cpuTotalCount;
            this.gpuAvailableCount = gpuAvailableCount;
            this.gpuTotalCount = gpuTotalCount;
            this.gpuTypes = gpuTypes;
            this.slaveCount = slaveCount;
            this.version = version;
        }

        @Nullable
        @JsonProperty("cluster_id")
        public String getId() {
            return id;
        }

        @Nullable
        @JsonProperty("comment")
        public String getComment() {
            return comment;
        }

        @JsonProperty("cpu_available")
        @Nullable
        public Integer getCpuAvailableCount() {
            return cpuAvailableCount;
        }

        @JsonProperty("cpu_count")
        @Nullable
        public Integer getCpuTotalCount() {
            return cpuTotalCount;
        }

        @JsonProperty("gpu_available")
        @Nullable
        public Integer getGpuAvailableCount() {
            return gpuAvailableCount;
        }

        @JsonProperty("gpu_count")
        @Nullable
        public Integer getGpuTotalCount() {
            return gpuTotalCount;
        }

        @JsonProperty("gpu_models")
        @Nullable
        public List<String> getGpuTypes() {
            return gpuTypes;
        }

        @JsonProperty("slave_count")
        @Nullable
        public Integer getSlaveCount() {
            return slaveCount;
        }

        @JsonProperty("version")
        @Nullable
        public String getVersion() {
            return version;
        }

        @Override
        public String toString() {
            return "GPULabCluster{" +
                    "id='" + id + '\'' +
                    ", comment='" + comment + '\'' +
                    ", cpuAvailableCount='" + cpuAvailableCount + '\'' +
                    ", cpuTotalCount='" + cpuTotalCount + '\'' +
                    ", gpuAvailableCount='" + gpuAvailableCount + '\'' +
                    ", gpuTotalCount='" + gpuTotalCount + '\'' +
                    ", gpuTypes=" + gpuTypes +
                    ", slaveCount=" + slaveCount +
                    ", version='" + version + '\'' +
                    '}';
        }
    }

    @JsonInclude(JsonInclude.Include.NON_NULL)
    @JsonIgnoreProperties(ignoreUnknown=true)
    private static class GPULabJob {
        @Nullable private final String uuid;
        @Nullable private final String created;
        @Nullable private final String version;
        @Nullable private final String name;
//        @Nullable private final String nvdocker_data;
        @Nullable private final String project;
        @Nullable private final String queueTime;
        @Nullable private final String state;
        @Nullable private final String username;
//        @Nullable private final Map run_details; //deadlineData endDate slaveDnsName slaveHostname startDate

        @JsonCreator
        public GPULabJob(@JsonProperty("uuid") @Nullable String uuid,
                         @JsonProperty("created") @Nullable String created,
                         @JsonProperty("version") @Nullable String version,
                         @JsonProperty("name") @Nullable String name,
                         @JsonProperty("project") @Nullable String project,
                         @JsonProperty("queue_time") @Nullable String queueTime,
                         @JsonProperty("state") @Nullable String state,
                         @JsonProperty("username") @Nullable String username) {
            this.uuid = uuid;
            this.created = created;
            this.version = version;
            this.name = name;
            this.project = project;
            this.queueTime = queueTime;
            this.state = state;
            this.username = username;
        }

        @JsonProperty
        @Nullable
        public String getUuid() {
            return uuid;
        }

        @JsonProperty
        @Nullable
        public String getCreated() {
            return created;
        }

        @JsonProperty
        @Nullable
        public String getVersion() {
            return version;
        }

        @JsonProperty
        @Nullable
        public String getName() {
            return name;
        }

        @JsonProperty
        @Nullable
        public String getProject() {
            return project;
        }

        @JsonProperty("queue_time")
        @Nullable
        public String getQueueTime() {
            return queueTime;
        }

        @JsonProperty
        @Nullable
        public String getState() {
            return state;
        }

        @JsonProperty
        @Nullable
        public String getUsername() {
            return username;
        }

        @Override
        public String toString() {
            return "GPULabJob{" +
                    "uuid='" + uuid + '\'' +
                    ", created='" + created + '\'' +
                    ", version='" + version + '\'' +
                    ", name='" + name + '\'' +
                    ", project='" + project + '\'' +
                    ", queueTime='" + queueTime + '\'' +
                    ", state='" + state + '\'' +
                    ", username='" + username + '\'' +
                    '}';
        }
    }

    @JsonInclude(JsonInclude.Include.NON_NULL)
    @JsonIgnoreProperties(ignoreUnknown=true)
    private static class GPULabSlave {
        @Nullable private final String version;
        @Nullable private final String slaveHostname;
        @Nullable private final String clusterId;
        @Nullable private final List<String> gpuModels;
        @Nullable private final List<String> cpuModels;
        @Nullable private final String workerCount;
        @Nullable private final String systemMemoryMb;
        @Nullable private final Integer gpuCount;
        @Nullable private final Integer cpuCount;
        @Nullable private final Integer workerInuse;
        @Nullable private final Integer systemMemoryInuseMb;
        @Nullable private final Integer gpuInuse;
        @Nullable private final Integer cpuInuse;
        @Nullable private final String cudaVersionFull;
        @Nullable private final String lastUpdate;
        @Nullable private final String comment;

        @JsonCreator
        public GPULabSlave(@JsonProperty("version")                @Nullable String version, 
                           @JsonProperty("slave_hostname")         @Nullable String slaveHostname, 
                           @JsonProperty("cluster_id")             @Nullable String clusterId, 
                           @JsonProperty("gpu_models")             @Nullable List<String> gpuModels, 
                           @JsonProperty("cpu_models")             @Nullable List<String> cpuModels,
                           @JsonProperty("worker_count")           @Nullable String workerCount,
                           @JsonProperty("system_memory_mb")       @Nullable String systemMemoryMb, 
                           @JsonProperty("gpu_count")              @Nullable Integer gpuCount, 
                           @JsonProperty("cpu_count")              @Nullable Integer cpuCount,
                           @JsonProperty("worker_inuse")           @Nullable Integer workerInuse, 
                           @JsonProperty("system_memory_inuse_mb") @Nullable Integer systemMemoryInuseMb, 
                           @JsonProperty("gpu_inuse")              @Nullable Integer gpuInuse, 
                           @JsonProperty("cpu_inuse")              @Nullable Integer cpuInuse, 
                           @JsonProperty("cuda_version_full")      @Nullable String cudaVersionFull, 
                           @JsonProperty("last_update")            @Nullable String lastUpdate, 
                           @JsonProperty("comment")                @Nullable String comment) {
            this.version = version;
            this.slaveHostname = slaveHostname;
            this.clusterId = clusterId;
            this.gpuModels = gpuModels;
            this.cpuModels = cpuModels;
            this.workerCount = workerCount;
            this.systemMemoryMb = systemMemoryMb;
            this.gpuCount = gpuCount;
            this.cpuCount = cpuCount;
            this.workerInuse = workerInuse;
            this.systemMemoryInuseMb = systemMemoryInuseMb;
            this.gpuInuse = gpuInuse;
            this.cpuInuse = cpuInuse;
            this.cudaVersionFull = cudaVersionFull;
            this.lastUpdate = lastUpdate;
            this.comment = comment;
        }

        @JsonProperty("version")
        @Nullable
        public String getVersion() {
            return version;
        }

        @JsonProperty("slave_hostname")
        @Nullable
        public String getSlaveHostname() {
            return slaveHostname;
        }

        @JsonProperty("cluster_id")
        @Nullable
        public String getClusterId() {
            return clusterId;
        }

        @JsonProperty("gpu_models")
        @Nullable
        public List<String> getGpuModels() {
            return gpuModels;
        }

        @JsonProperty("cpu_models")
        @Nullable
        public List<String> getCpuModels() {
            return cpuModels;
        }

        @JsonProperty("worker_count")
        @Nullable
        public String getWorkerCount() {
            return workerCount;
        }

        @JsonProperty("system_memory_mb")
        @Nullable
        public String getSystemMemoryMb() {
            return systemMemoryMb;
        }

        @JsonProperty("gpu_count")
        @Nullable
        public Integer getGpuCount() {
            return gpuCount;
        }

        @JsonProperty("cpu_count")
        @Nullable
        public Integer getCpuCount() {
            return cpuCount;
        }

        @JsonProperty("worker_inuse")
        @Nullable
        public Integer getWorkerInuse() {
            return workerInuse;
        }

        @JsonProperty("system_memory_inuse_mb")
        @Nullable
        public Integer getSystemMemoryInuseMb() {
            return systemMemoryInuseMb;
        }

        @JsonProperty("gpu_inuse")
        @Nullable
        public Integer getGpuInuse() {
            return gpuInuse;
        }

        @JsonProperty("cpu_inuse")
        @Nullable
        public Integer getCpuInuse() {
            return cpuInuse;
        }

        @JsonProperty("cuda_version_full")
        @Nullable
        public String getCudaVersionFull() {
            return cudaVersionFull;
        }

        @JsonProperty("last_update")
        @Nullable
        public String getLastUpdate() {
            return lastUpdate;
        }

        @JsonProperty("comment")
        @Nullable
        public String getComment() {
            return comment;
        }

        @Override
        public String toString() {
            return "GPULabSlave{" +
                    "version='" + version + '\'' +
                    ", slaveHostname='" + slaveHostname + '\'' +
                    ", clusterId='" + clusterId + '\'' +
                    ", gpuModels=" + gpuModels +
                    ", cpuModels=" + cpuModels +
                    ", workerCount='" + workerCount + '\'' +
                    ", systemMemoryMb='" + systemMemoryMb + '\'' +
                    ", gpuCount=" + gpuCount +
                    ", cpuCount=" + cpuCount +
                    ", workerInuse=" + workerInuse +
                    ", systemMemoryInuseMb=" + systemMemoryInuseMb +
                    ", gpuInuse=" + gpuInuse +
                    ", cpuInuse=" + cpuInuse +
                    ", cudaVersionFull='" + cudaVersionFull + '\'' +
                    ", lastUpdate='" + lastUpdate + '\'' +
                    ", comment='" + comment + '\'' +
                    '}';
        }
    }
}
